@@ -4,8 +4,6 @@ import Data.Either
 import qualified Data.ByteString.Lazy as L
 import qualified Data.ByteString.Lazy.Char8 as LC
 import qualified Data.ByteString as B
-import qualified Data.Text.Lazy as T
-import Data.Text.Lazy.Encoding (decodeUtf8)
 import Data.X509
 import qualified Data.X509 as X509
 import Data.List (find)
@@ -19,18 +17,11 @@ import System.Exit
 import System.X509
 import Data.X509.CertificateStore
 import Data.X509.Validation
+import Data.Hourglass
 
 -- for signing/verifying certificate
-import qualified Crypto.Hash.SHA1 as SHA1
-import qualified Crypto.Hash.SHA224 as SHA224
-import qualified Crypto.Hash.SHA256 as SHA256
-import qualified Crypto.Hash.SHA512 as SHA512
-import qualified Crypto.Hash.MD2 as MD2
-import qualified Crypto.Hash.MD5 as MD5
-import qualified Crypto.PubKey.HashDescr as HD
+import Crypto.Hash
 import qualified Crypto.PubKey.RSA as RSA
-import qualified Crypto.Types.PubKey.RSA as RSA
-import qualified Crypto.Types.PubKey.DSA as DSA
 import qualified Crypto.PubKey.RSA.PKCS15 as RSA
 import qualified Crypto.PubKey.DSA as DSA
 
@@ -38,9 +29,13 @@ import Data.ASN1.Encoding
 import Data.ASN1.BinaryEncoding
 import Data.ASN1.Types
 import Data.ASN1.BitArray
+import Data.X509.Memory
 import Text.Printf
 import Numeric
 
+formatValidity (start,end) = p start ++ " to " ++ p end
+  where p t = timePrint ("YYYY-MM-DD H:MI:S" :: String) t
+
 hexdump :: B.ByteString -> String
 hexdump bs = concatMap hex $ B.unpack bs
     where hex n
@@ -62,18 +57,35 @@ showExts es@(Extensions Nothing) = do
 showExts es@(Extensions (Just exts)) = do
     mapM_ showExt exts
     putStrLn "known extensions decoded: "
-    showKnownExtension (X509.extensionGet es :: Maybe X509.ExtBasicConstraints)
-    showKnownExtension (X509.extensionGet es :: Maybe X509.ExtKeyUsage)
-    showKnownExtension (X509.extensionGet es :: Maybe X509.ExtExtendedKeyUsage)
-    showKnownExtension (X509.extensionGet es :: Maybe X509.ExtSubjectKeyId)
-    showKnownExtension (X509.extensionGet es :: Maybe X509.ExtSubjectAltName)
-    showKnownExtension (X509.extensionGet es :: Maybe X509.ExtAuthorityKeyId)
+    showKnownExtension "basic-constraint" (X509.extensionGetE es :: Maybe (Either String X509.ExtBasicConstraints))
+    showKnownExtension "key-usage" (X509.extensionGetE es :: Maybe (Either String X509.ExtKeyUsage))
+    showKnownExtension "extended-key-usage" (X509.extensionGetE es :: Maybe (Either String X509.ExtExtendedKeyUsage))
+    showKnownExtension "subject-key-id" (X509.extensionGetE es :: Maybe (Either String X509.ExtSubjectKeyId))
+    showKnownExtension "subject-alt-name" (X509.extensionGetE es :: Maybe (Either String X509.ExtSubjectAltName))
+    showKnownExtension "authority-key-id" (X509.extensionGetE es :: Maybe (Either String X509.ExtAuthorityKeyId))
     where
         showExt (ExtensionRaw oid critical asn1) = do
             putStrLn ("  OID:  " ++ show oid ++ " critical: " ++ show critical)
-            putStrLn ("        " ++ show asn1)
-        showKnownExtension Nothing  = return ()
-        showKnownExtension (Just e) = putStrLn ("  " ++ show e)
+            showASN1 8 asn1
+        showKnownExtension _ Nothing  = return ()
+        showKnownExtension n (Just (Left e)) = putStrLn ("  " ++ n ++ ": ERROR: " ++ show e)
+        showKnownExtension _ (Just (Right e)) = putStrLn ("  " ++ show e)
+
+showCertSmall :: SignedCertificate -> IO ()
+showCertSmall signedCert = do
+    putStrLn "subject: "
+    showDN $ X509.certSubjectDN cert
+    putStrLn ("valid:  " ++ formatValidity (X509.certValidity cert))
+    case X509.certPubKey cert of
+        X509.PubKeyRSA pubkey     -> printf "public key: RSA (%d bits)\n" (RSA.public_size pubkey * 8)
+        X509.PubKeyDSA pubkey     -> printf "public key: DSA\n"
+        X509.PubKeyUnknown oid ws -> printf "public key: unknown: %s\n" (show oid)
+        pk                        -> printf "public key: %s\n" (show pk)
+  where
+    signed  = X509.getSigned signedCert
+    --sigalg  = X509.signedAlg signed
+    --sigbits = X509.signedSignature signed
+    cert    = X509.signedObject signed
 
 showCert :: SignedCertificate -> IO ()
 showCert signedCert = do
@@ -84,7 +96,7 @@ showCert signedCert = do
     showDN $ X509.certIssuerDN cert
     putStrLn "subject:"
     showDN $ X509.certSubjectDN cert
-    putStrLn ("valid:  " ++ show (X509.certValidity cert))
+    putStrLn ("valid:  " ++ formatValidity (X509.certValidity cert))
     case X509.certPubKey cert of
         X509.PubKeyRSA pubkey -> do
             putStrLn "public key RSA:"
@@ -115,12 +127,12 @@ showCert signedCert = do
     cert    = X509.signedObject signed
 
 
-showRSAKey :: (RSA.KeyPair) -> String
-showRSAKey (RSA.KeyPair privkey) = unlines
+showRSAKey :: RSA.PrivateKey -> String
+showRSAKey privkey = unlines
     [ "len-modulus:      " ++ (show $ RSA.public_size pubkey)
     , "modulus:          " ++ (show $ RSA.public_n pubkey)
-    , "public exponant:  " ++ (show $ RSA.public_e pubkey)
-    , "private exponant: " ++ (show $ RSA.private_d privkey)
+    , "public exponent:  " ++ (show $ RSA.public_e pubkey)
+    , "private exponent: " ++ (show $ RSA.private_d privkey)
     , "p1:               " ++ (show $ RSA.private_p privkey)
     , "p2:               " ++ (show $ RSA.private_q privkey)
     , "exp1:             " ++ (show $ RSA.private_dP privkey)
@@ -129,17 +141,17 @@ showRSAKey (RSA.KeyPair privkey) = unlines
     ]
   where pubkey = RSA.private_pub privkey
 
-showDSAKey :: DSA.KeyPair -> String
-showDSAKey (DSA.KeyPair params pubnum privnum) = unlines
+showDSAKey :: DSA.PrivateKey -> String
+showDSAKey (DSA.PrivateKey params privnum) = unlines
     [ "priv     " ++ (printf "%x" $ privnum)
-    , "pub:     " ++ (printf "%x" $ pubnum)
     , "p:       " ++ (printf "%x" $ DSA.params_p params)
     , "q:       " ++ (printf "%x" $ DSA.params_q params)
     , "g:       " ++ (printf "%x" $ DSA.params_g params)
     ]
 
 showASN1 :: Int -> [ASN1] -> IO ()
-showASN1 at = prettyPrint at where
+showASN1 at = prettyPrint at
+  where
     indent n = putStr (replicate n ' ')
 
     prettyPrint n []                 = return ()
@@ -155,12 +167,12 @@ showASN1 at = prettyPrint at where
     p (OID is)               = putStr ("OID: " ++ show is)
     p (Real d)               = putStr "real"
     p (Enumerated _)         = putStr "enum"
-    p (Start Sequence)       = putStr "sequence"
-    p (End Sequence)         = putStr "end-sequence"
-    p (Start Set)            = putStr "set"
-    p (End Set)              = putStr "end-set"
-    p (Start _)              = putStr "container"
-    p (End _)                = putStr "end-container"
+    p (Start Sequence)       = putStr "{"
+    p (End Sequence)         = putStr "}"
+    p (Start Set)            = putStr "["
+    p (End Set)              = putStr "]"
+    p (Start (Container x y)) = putStr ("< " ++ show x ++ " " ++ show y)
+    p (End (Container x y))   = putStr ("> " ++ show x ++ " " ++ show y)
     p (ASN1String cs)        = putCS cs
     p (ASN1Time TimeUTC time tz)      = putStr ("utctime: " ++ show time)
     p (ASN1Time TimeGeneralized time tz) = putStr ("generalizedtime: " ++ show time)
@@ -169,9 +181,9 @@ showASN1 at = prettyPrint at where
     putCS (ASN1CharacterString UTF8 t)         = putStr ("utf8string:" ++ show t)
     putCS (ASN1CharacterString Numeric bs)     = putStr "numericstring:"
     putCS (ASN1CharacterString Printable t)    = putStr ("printablestring: " ++ show t)
-    putCS (ASN1CharacterString T61 bs)         = putStr "t61string:"
+    putCS (ASN1CharacterString T61 bs)         = putStr ("t61string:" ++ show bs)
     putCS (ASN1CharacterString VideoTex bs)    = putStr "videotexstring:"
-    putCS (ASN1CharacterString IA5 bs)         = putStr "ia5string:"
+    putCS (ASN1CharacterString IA5 bs)         = putStr ("ia5string:" ++ show bs)
     putCS (ASN1CharacterString Graphic bs)     = putStr "graphicstring:"
     putCS (ASN1CharacterString Visible bs)     = putStr "visiblestring:"
     putCS (ASN1CharacterString General bs)     = putStr "generalstring:"
@@ -248,17 +260,20 @@ doKeyMain files = do
     pems <- readPEMFile (head files)
     forM_ pems $ \pem -> do
         let content = either (error . show) id $ decodeASN1' BER (pemContent pem)
-        case pemName pem of
-            "RSA PRIVATE KEY" ->
-                case fromASN1 content of
-                    Left err    -> error ("not a valid RSA key: " ++ err)
-                    Right (k,_) -> putStrLn "RSA KEY" >> putStrLn (showRSAKey k)
-            "DSA PRIVATE KEY" ->
-                case fromASN1 content of
-                    Left err    -> error ("not a valid DSA key: " ++ err)
-                    Right (k,_) -> putStrLn "DSA KEY" >> putStrLn (showDSAKey k)
-            _                 ->
-                putStrLn ("unknown private key: " ++ show (pemName pem))
+            privkey = pemToKey [] pem
+        case privkey of
+            [Just (X509.PrivKeyRSA k)] ->
+                putStrLn "RSA KEY" >> putStrLn (showRSAKey k)
+            [Just (X509.PrivKeyDSA k)] ->
+                putStrLn "DSA KEY" >> putStrLn (showDSAKey k)
+            _ -> error "private key unknown"
+
+doSystemMain _ = do
+    store <- getSystemCertificateStore
+    let certs = listCertificates store
+    mapM_ showCertSmall certs
+    putStrLn $ replicate 72 '='
+    putStrLn $ show (length certs) ++ " certificates loaded"
 
 optionsCert =
     [ Option []     ["hash"] (NoArg ShowHash) "output certificate hash"
@@ -272,6 +287,9 @@ crlMain = getoptMain [] $ \o n -> doCRLMain o n
 keyMain = getoptMain [] $ \o n -> doKeyMain n
 asn1Main = getoptMain [] $ \o n -> doASN1Main n
 
+systemMain = getoptMain [] $ \o n -> doSystemMain n
+
+
 getoptMain :: [OptDescr a] -> ([a] -> [String] -> IO ()) -> [String] -> IO ()
 getoptMain opts f as =
     case getOpt Permute opts as of
@@ -284,6 +302,7 @@ usage = do
     putStrLn "  cert: process X509 certificate"
     putStrLn "  crl : process CRL certificate"
     putStrLn "  asn1: show file asn1"
+    putStrLn "  system: show system certificates"
 
 main = do
     args <- getArgs
@@ -294,4 +313,5 @@ main = do
         "key":as  -> keyMain as
         "crl":as  -> crlMain as
         "asn1":as -> asn1Main as
+        "system":as -> systemMain as
         _         -> usage
